from __future__ import annotations

from typing import Any, Dict, Literal

import pandas as pd
from sklearn import model_selection
from sklearn.model_selection._split import BaseCrossValidator


def create_folds(
    df: pd.DataFrame,
    *,
    resample_strategy: str,
    resample_params: Dict[str, Any],
    group_by: str | None = None,
    stratify_by: str | None = None,
    fold_column: Literal["fold"] = "fold",
) -> pd.DataFrame:
    """
    Assign fold numbers to rows in a DataFrame based on specified resampling
    strategy.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to which the folds will be assigned.
    resample_strategy : str
        The resampling strategy, corresponds to sklearn's model_selection methods.
    resample_params : dict
        Parameters to pass to the resampling strategy constructor.
    group_by : str or None, optional
        Column to group data before splitting, by default None.
    stratify_by : str or None, optional
        Column to stratify split, by default None.
    fold_column : str, optional
        Name of the column to store fold numbers, by default "fold".

    Returns
    -------
    pd.DataFrame
        DataFrame with an additional column indicating fold numbers.

    Notes
    -----
    Omit the use of `train_test_split` since the same result can be achieved by
    using `(Stratified)(Group)KFold` with `n_splits=2`.
    """
    cv: BaseCrossValidator = getattr(model_selection, resample_strategy)(**resample_params)
    stratify = df[stratify_by].values if stratify_by else None
    groups = df[group_by].values if group_by else None

    for _fold, (_train_idx, valid_idx) in enumerate(cv.split(df, stratify, groups)):
        df.loc[valid_idx, fold_column] = _fold
    df[fold_column] = df[fold_column].astype(int)
    return df



from matplotlib_inline import backend_inline  # type: ignore


def use_svg_display() -> None:
    """Use the svg format to display a plot in Jupyter."""
    backend_inline.set_matplotlib_formats("svg")



from collections.abc import MutableSequence
from typing import Generic, Iterable, Optional

from omnivault._types._generic import T


class Deque(MutableSequence[T], Generic[T]):
    def __init__(self, iterable: Optional[Iterable[T]] = None, maxlen: Optional[int] = None):
        self.maxlen = maxlen
        self._data = list(iterable) if iterable is not None else []

        if self.maxlen is not None and len(self._data) > self.maxlen:
            self._data = self._data[-self.maxlen :]

    def append(self, item: T) -> None:
        if self.maxlen is None or len(self._data) < self.maxlen:
            self._data.append(item)
        else:
            self._data.pop(0)
            self._data.append(item)

    def appendleft(self, item: T) -> None:
        if self.maxlen is None or len(self._data) < self.maxlen:
            self._data.insert(0, item)
        else:
            self._data.pop()
            self._data.insert(0, item)

    def pop(self) -> T:
        return self._data.pop()

    def popleft(self) -> T:
        return self._data.pop(0)

    def __len__(self) -> int:
        return len(self._data)

    def __getitem__(self, index) -> T:
        return self._data[index]

    def __setitem__(self, index, value: T) -> None:
        self._data[index] = value

    def __delitem__(self, index) -> None:
        del self._data[index]

    def insert(self, index: int, value: T) -> None:
        if self.maxlen is None or len(self._data) < self.maxlen:
            self._data.insert(index, value)
        else:
            raise OverflowError("Deque is at its maximum size")


# Example usage
my_deque = Deque([1, 2, 3], maxlen=3)
my_deque.append(4)
print(my_deque)  # Output might be [2, 3, 4] if maxlen is reached



from typing import Dict, List

import numpy as np


def calculate_statistics(data: List[float], suffix: str = "") -> Dict[str, float]:
    """
    Calculate statistics for a list of data.

    Parameters
    ----------
    data : List[float]
        List of 1D data such as a list of time taken in seconds on each rank.

    Returns
    -------
    Dict[str, float]
        A dictionary containing the following statistics:
        - mean: The mean of the data.
        - median: The median of the data.
        - variance: The variance of the data.
        - standard_deviation: The standard deviation of the data.
    """
    mean = sum(data) / len(data)
    median = float(np.median(data))
    standard_deviation = float(np.std(data))
    variance = float(np.var(data))
    min_value, max_value = min(data), max(data)
    total = sum(data)

    return {
        "mean" + suffix: mean,
        "median" + suffix: median,
        "variance" + suffix: variance,
        "standard_deviation" + suffix: standard_deviation,
        "min" + suffix: min_value,
        "max" + suffix: max_value,
        "total" + suffix: total,
    }



from typing import Any, Callable, List

from IPython.core.display import HTML, display_html


# mypy: disable-error-code="no-untyped-call"
class TestFramework:
    """
    A simple testing framework for executing tests with descriptions.

    Attributes
    ----------
    tests: (List[Callable])
        A list to store the test functions.
    """

    def __init__(self) -> None:
        """Initialize the TestFramework object."""
        self.tests: List[Callable[[], None]] = []

    def describe(self, description: str) -> Callable[[Callable[[], None]], None]:
        """
        Decorator for describing a test.

        Parameters
        ----------
        description: str
            The description of the test.

        Returns
        -------
        wrapper: Callable
            A wrapped function with a description.
        """

        def wrapper(func: Callable[[], None]) -> None:
            """Execute the function and display its description."""
            display_html(HTML(f"<span style='color: blue;'>Description: {description}</span>"))
            func()

        return wrapper

    def individual_test(self, description: str) -> Callable[[Callable[[], None]], None]:
        """
        Decorator for running a single test.

        Parameters
        ----------
        description: str
            Description of what the test should do.

        Returns
        -------
        wrapper: Callable
            A wrapped function that runs the test.
        """

        def wrapper(func: Callable[[], None]) -> None:
            """Execute the function and display the test result."""
            try:
                func()
                display_html(HTML(f"<span style='color: green;'>  [Pass]</span> {description}"))
            except AssertionError as e:
                display_html(HTML(f"<span style='color: red;'>  [Fail]</span> {description} - {e}"))

        return wrapper

    def assert_equals(self, actual: Any, expected: Any, message: str) -> None:
        """
        Assert that the actual and expected values are equal.

        Parameters
        ----------
        actual  : Any
            The actual value.
        expected: Any
            The expected value.
        message : str
            The message to display if the assertion fails.

        Raises
        ------
        AssertionError: If the actual and expected values are not equal.
        """
        try:
            assert actual == expected, message
        except AssertionError as err:
            message = f"Test failed: {message}\nExpected: {expected}, but got: {actual}"
            raise AssertionError(message) from err



"""
The context module in the omnivault.dsa.searching_algorithms package. It defines
the SearchContext class, which acts as the primary interface for executing
search strategies in the Strategy Design Pattern.

The SearchContext class holds a reference to a search strategy and delegates the
searching task to it. This allows for dynamic swapping of search algorithms at
runtime, adhering to the principles of the Strategy Pattern.
"""
from typing import Sequence

from omnivault._types._generic import Real
from omnivault.dsa.searching_algorithms.base import Search


class SearchContext:
    """The Context defines the interface of interest to clients."""

    def __init__(self, strategy: Search) -> None:
        """The context defines the interface of interest to clients and accepts
        a strategy through the constructor."""
        self._strategy = strategy

    @property
    def strategy(self) -> Search:
        """The Context maintains a reference to one of the Strategy objects. The
        Context does not know the concrete class of a strategy. It should work
        with all strategies via the Strategy interface."""
        return self._strategy

    @strategy.setter
    def strategy(self, strategy: Search) -> None:
        """Usually, the Context allows replacing a Strategy object at runtime."""
        self._strategy = strategy

    def execute_search(self, container: Sequence[Real], target: Real) -> int:
        """Here the context delegates some work to the strategy object instead
        of implementing multiple versions of the algorithm on its own."""
        return self.strategy.search(container, target)



"""
`NewType` is used to create distinct types for type checkers. It is helpful when
you want to differentiate between different kinds of data that are structurally
the same. For instance, if you want to distinguish between a regular integer and
a non-negative integer at the type-checking level, `NewType` would be
appropriate.

However, `NewType` does not enforce any runtime checks. It's essentially a cast
for the type checker, and at runtime, it's equivalent to the base type.
"""

from typing import NewType

# fmt: off
NonNegativeInt = NewType("NonNegativeInt", int) # Too strong typing lol need to cast everywhere, may not be worth it.
PositiveInt    = NewType("PositiveInt", int)
# fmt: on



from typing import Any, Dict, List, Type, Union

from pydantic import BaseModel, Field, field_validator

from omnivault._types._alias import NotGiven


class DataConfig(BaseModel):
    """The data config."""

    context_length: int = Field(
        default=128,
        description="The context length depends on how we tokenize, whether on a character level or word level.",
    )
    dataset_name: str = Field(
        default=None,
        description="The name of the dataset. Also the stem of the url or data path, for example, if the filepath is `data/abc.txt`, then the dataset name is `abc`.",
    )
    dataset_size: Union[int, None] = Field(default=2, description="The size of the dataset.")
    dataset_path: str = Field(default=None, description="The path to the dataset.")
    dataset_dir: str = Field(default=None, description="The directory to the dataset.")
    dataset_url: str = Field(default=None, description="The url to the dataset.")

    split: Union[List[float], None] = Field(default=[0.7, 0.1, 0.2], description="The split ratio of the dataset.")

    collate_fn: Union[Dict[str, Any], NotGiven, None] = Field(
        default={
            "batch_first": True,
            "pad_token_id": 16,
        },  # TODO: `pad_token_id` should be interpolated from `MaybeConstant`.
        description="The collate function config.",
    )

    train_loader: Dict[str, Any] = Field(
        default={"batch_size": 32, "shuffle": True, "num_workers": 0, "pin_memory": False, "drop_last": False},
        description="The train loader config.",
    )
    valid_loader: Union[Dict[str, Any], None] = Field(
        default={"batch_size": 32, "shuffle": False, "num_workers": 0, "pin_memory": False, "drop_last": False},
        description="The validation loader config.",
    )
    test_loader: Union[Dict[str, Any], None] = Field(
        default={"batch_size": 32, "shuffle": False, "num_workers": 0, "pin_memory": False, "drop_last": False},
        description="The test loader config.",
    )

    class Config:
        arbitrary_types_allowed = True

    # FIXME: hard to handle since collate_fn can be NotGiven but yaml config can only indicate null.
    # unless I do __target__?
    @field_validator("collate_fn")
    def coerce_collate_fn(cls: Type["DataConfig"], v: Union[Dict[str, Any], NotGiven, None]) -> Dict[str, Any]:  # type: ignore
        if v is None:
            return {}
        if isinstance(v, NotGiven):
            return {}
        else:
            return v



from dataclasses import dataclass, field
from typing import Dict, List

import torch

from omnivault.transformer.config.composer import DataConfig
from omnivault.transformer.config.constants import MaybeConstant
from omnivault.transformer.core.dataset import AdderDataset, AdderDatasetYield
from omnivault.transformer.core.tokenizer import AdderTokenizer
from omnivault.transformer.core.vocabulary import AdderVocabulary
from omnivault.transformer.utils.general_utils import download_and_read_sequences


@dataclass
class AdderGroundTruth:
    token_to_index: Dict[str, int] = field(  # TODO: consider using Literal
        default_factory=lambda: {
            "0": 0,
            "1": 1,
            "2": 2,
            "3": 3,
            "4": 4,
            "5": 5,
            "6": 6,
            "7": 7,
            "8": 8,
            "9": 9,
            "+": 10,
            "*": 11,
            "-": 12,
            "=": 13,
            "<BOS>": 14,
            "<EOS>": 15,
            "<PAD>": 16,
            "<UNK>": 17,
        }
    )
    index_to_token: Dict[int, str] = field(
        default_factory=lambda: {
            0: "0",
            1: "1",
            2: "2",
            3: "3",
            4: "4",
            5: "5",
            6: "6",
            7: "7",
            8: "8",
            9: "9",
            10: "+",
            11: "*",
            12: "-",
            13: "=",
            14: "<BOS>",
            15: "<EOS>",
            16: "<PAD>",
            17: "<UNK>",
        }
    )

    # test bad sequences like 01+02=03?
    seq_len: int = 10  # all sequences are padded to this length in this test example

    sequences: List[str] = field(default_factory=lambda: ["15+57=072", "01+02=003"])
    tokenized_sequences: List[List[str]] = field(
        default_factory=lambda: [
            ["<BOS>", "1", "5", "+", "5", "7", "=", "0", "7", "2", "<EOS>"],
            ["<BOS>", "0", "1", "+", "0", "2", "=", "0", "0", "3", "<EOS>"],
        ]
    )
    encoded_sequences: List[List[int]] = field(
        default_factory=lambda: [
            [14, 1, 5, 10, 5, 7, 13, 0, 7, 2, 15],
            [14, 0, 1, 10, 0, 2, 13, 0, 0, 3, 15],
        ]
    )
    decoded_sequences: List[str] = field(default_factory=lambda: ["15+57=072", "01+02=003"])

    inputs: List[torch.LongTensor] = field(
        default_factory=lambda: [
            torch.LongTensor([14, 1, 5, 10, 5, 7, 13, 0, 7, 2]),
            torch.LongTensor([14, 0, 1, 10, 0, 2, 13, 0, 0, 3]),
        ]
    )
    targets: List[torch.LongTensor] = field(
        default_factory=lambda: [
            torch.LongTensor([16, 16, 16, 16, 16, 16, 0, 7, 2, 15]),
            torch.LongTensor([16, 16, 16, 16, 16, 16, 0, 0, 3, 15]),
        ]
    )
    padding_masks: List[torch.BoolTensor] = field(
        default_factory=lambda: [
            torch.BoolTensor([True, True, True, True, True, True, True, True, True, True]),
            torch.BoolTensor([True, True, True, True, True, True, True, True, True, True]),
        ]
    )
    future_masks: List[torch.BoolTensor] = field(
        default_factory=lambda: [
            torch.BoolTensor(
                [
                    [True, False, False, False, False, False, False, False, False, False],
                    [True, True, False, False, False, False, False, False, False, False],
                    [True, True, True, False, False, False, False, False, False, False],
                    [True, True, True, True, False, False, False, False, False, False],
                    [True, True, True, True, True, False, False, False, False, False],
                    [True, True, True, True, True, True, False, False, False, False],
                    [True, True, True, True, True, True, True, False, False, False],
                    [True, True, True, True, True, True, True, True, False, False],
                    [True, True, True, True, True, True, True, True, True, False],
                    [True, True, True, True, True, True, True, True, True, True],
                ]
            ),
            torch.BoolTensor(
                [
                    [True, False, False, False, False, False, False, False, False, False],
                    [True, True, False, False, False, False, False, False, False, False],
                    [True, True, True, False, False, False, False, False, False, False],
                    [True, True, True, True, False, False, False, False, False, False],
                    [True, True, True, True, True, False, False, False, False, False],
                    [True, True, True, True, True, True, False, False, False, False],
                    [True, True, True, True, True, True, True, False, False, False],
                    [True, True, True, True, True, True, True, True, False, False],
                    [True, True, True, True, True, True, True, True, True, False],
                    [True, True, True, True, True, True, True, True, True, True],
                ]
            ),
        ]
    )

    mock_batch: List[AdderDatasetYield] = field(init=False)

    inputs_collated: torch.LongTensor = field(
        default_factory=lambda: torch.LongTensor([[14, 1, 5, 10, 5, 7, 13, 0, 7, 2], [14, 0, 1, 10, 0, 2, 13, 0, 0, 3]])
    )
    targets_collated: torch.LongTensor = field(
        default_factory=lambda: torch.LongTensor(
            [
                [16, 16, 16, 16, 16, 16, 0, 7, 2, 15],
                [16, 16, 16, 16, 16, 16, 0, 0, 3, 15],
            ]
        )
    )

    padding_masks_collated: torch.BoolTensor = field(
        default_factory=lambda: torch.BoolTensor(
            [
                [
                    [
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                    ]
                ],
                [
                    [
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True, True],
                    ]
                ],
            ]
        )
    )
    # shape: (batch_size, 1, seq_len, seq_len)
    future_masks_collated: torch.BoolTensor = field(
        default_factory=lambda: torch.BoolTensor(
            [
                [
                    [
                        [True, False, False, False, False, False, False, False, False, False],
                        [True, True, False, False, False, False, False, False, False, False],
                        [True, True, True, False, False, False, False, False, False, False],
                        [True, True, True, True, False, False, False, False, False, False],
                        [True, True, True, True, True, False, False, False, False, False],
                        [True, True, True, True, True, True, False, False, False, False],
                        [True, True, True, True, True, True, True, False, False, False],
                        [True, True, True, True, True, True, True, True, False, False],
                        [True, True, True, True, True, True, True, True, True, False],
                        [True, True, True, True, True, True, True, True, True, True],
                    ]
                ],
                [
                    [
                        [True, False, False, False, False, False, False, False, False, False],
                        [True, True, False, False, False, False, False, False, False, False],
                        [True, True, True, False, False, False, False, False, False, False],
                        [True, True, True, True, False, False, False, False, False, False],
                        [True, True, True, True, True, False, False, False, False, False],
                        [True, True, True, True, True, True, False, False, False, False],
                        [True, True, True, True, True, True, True, False, False, False],
                        [True, True, True, True, True, True, True, True, False, False],
                        [True, True, True, True, True, True, True, True, True, False],
                        [True, True, True, True, True, True, True, True, True, True],
                    ]
                ],
            ]
        )
    )

    def __post_init__(self) -> None:
        # Construct mock_batch using the existing fields
        self.mock_batch = [
            (self.inputs[i], self.targets[i], self.padding_masks[i], self.future_masks[i])
            for i in range(len(self.inputs))
        ]


ADDER_GROUND_TRUTH = AdderGroundTruth()

adder_mock_batch_ = ADDER_GROUND_TRUTH.mock_batch

constants = MaybeConstant(
    NUM_DIGITS=2,
    TOKENS=["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "+", "*", "-", "=", "<BOS>", "<EOS>", "<PAD>", "<UNK>"],
)
adder_vocab_ = AdderVocabulary.from_tokens(tokens=constants.TOKENS, num_digits=constants.NUM_DIGITS)  # type: ignore[attr-defined]
adder_tokenizer_ = AdderTokenizer(vocabulary=adder_vocab_)

data = DataConfig(
    context_length=11,
    dataset_name="adder_dataset",
    dataset_size=10000,
    dataset_path="./data/adder/adder_dataset.txt",
    dataset_dir="./data/adder",
    dataset_url="https://raw.githubusercontent.com/gao-hongnan/omniverse/dev/omnivault/transformer/projects/adder/assets/adder_dataset.txt",
    split=[0.7, 0.2, 0.1],
    collate_fn={"batch_first": True, "pad_token_id": 16},
    train_loader={"batch_size": 256, "shuffle": True, "num_workers": 0, "pin_memory": False, "drop_last": False},
    valid_loader={"batch_size": 256, "shuffle": False, "num_workers": 0, "pin_memory": False, "drop_last": False},
    test_loader={"batch_size": 128, "shuffle": False, "num_workers": 0, "pin_memory": False, "drop_last": False},
)


sequences = list(download_and_read_sequences(url=data.dataset_url, dataset_name=data.dataset_name))
adder_dataset_ = AdderDataset(
    data=sequences, tokenizer=adder_tokenizer_
)  # NOTE: for end2end test, so load a small dataset here.
adder_mock_dataset_ = AdderDataset(data=ADDER_GROUND_TRUTH.sequences, tokenizer=adder_tokenizer_)  # type: ignore[arg-type]



from __future__ import annotations

from typing import Dict, List, Literal, Set, Tuple, Type

import torch
from torch import nn

from omnivault.transformer.modules.layers.normalization import LayerNorm


# FIXME: I have my own implementation of AddNorm, should I blacklist it or not?
def apply_weight_decay_to_different_param_groups(
    model: nn.Module, weight_decay: float
) -> List[Dict[Literal["params", "weight_decay"], List[torch.nn.Parameter] | float]]:
    """
    Categorizes parameters of a PyTorch model into two groups based on weight decay.

    Parameters that typically undergo weight decay (like Linear layer weights)
    and those that don't (like biases and LayerNorm weights) are separated.
    This categorization is useful for optimizers that apply weight decay differently
    to different types of parameters.

    Reference
    ----------
    PyTorch mingpt: https://github.com/pytorch/examples/blob/main/distributed/minGPT-ddp/mingpt/model.py

    Parameters
    ----------
    model : nn.Module
        The PyTorch model whose parameters are to be categorized.
    weight_decay : float
        The weight decay to be applied to the parameters that undergo weight decay.

    Returns
    -------
    optim_groups: List[Dict[str, torch.nn.Parameter]]
        A list containing two dictionaries, one for parameters to decay and
        one for parameters not to decay.

    Raises
    ------
    AssertionError
        If any parameter is found in both decay and no_decay sets.
    """

    decay: Set[str] = set()
    no_decay: Set[str] = set()
    whitelist_weight_modules: Tuple[Type[nn.Module], ...] = (nn.Linear,)
    blacklist_weight_modules: Tuple[Type[nn.Module], ...] = (nn.LayerNorm, nn.Embedding, LayerNorm)

    for module_name, module in model.named_modules():
        for parameter_name, _parameter in module.named_parameters():
            full_parameter_name = f"{module_name}.{parameter_name}" if module_name else parameter_name
            if parameter_name.endswith("bias"):
                # biases of all modules are not decayed
                no_decay.add(full_parameter_name)
            elif parameter_name.endswith("weight") and isinstance(module, whitelist_weight_modules):
                # weights of whitelisted modules are decayed
                decay.add(full_parameter_name)
            elif parameter_name.endswith("in_proj_weight"):
                # MHA projection layer, does not exist in my implementation
                decay.add(full_parameter_name)
            elif parameter_name.endswith("weight") and isinstance(module, blacklist_weight_modules):
                # weights of blacklisted modules are not decayed
                no_decay.add(full_parameter_name)
            elif (parameter_name.endswith("gamma") or parameter_name.endswith("beta")) and isinstance(
                module, LayerNorm
            ):
                # weights of LayerNorm modules are not decayed
                # TODO: why do I need to do this is because my custom LayerNorm has gamma and beta
                # as their "weight" and "bias" attributes, respectively.
                no_decay.add(full_parameter_name)
            elif parameter_name.endswith("pos_embed"):
                no_decay.add(full_parameter_name)

    param_dict = {parameter_name: parameter for parameter_name, parameter in model.named_parameters()}  # noqa: C416
    inter_params = decay & no_decay
    union_params = decay | no_decay
    assert not inter_params, f"Parameters {inter_params} are in both decay and no_decay sets."
    assert not (
        param_dict.keys() - union_params
    ), f"Parameters {param_dict.keys() - union_params} were not categorized."

    optim_groups: List[Dict[Literal["params", "weight_decay"], List[torch.nn.Parameter] | float]] = [
        {"params": [param_dict[parameter_name] for parameter_name in sorted(decay)], "weight_decay": weight_decay},
        {"params": [param_dict[parameter_name] for parameter_name in sorted(no_decay)], "weight_decay": 0.0},
    ]

    return optim_groups



"""State...Metadata...See how composer does it, quite elegant I'd say."""
from __future__ import annotations

from typing import Any, Dict, List, Type, Union

import torch
from pydantic import BaseModel, Field
from rich.pretty import pprint
from torch import nn
from torch.nn.parallel import DistributedDataParallel

from omnivault.utils.torch_utils.model_utils import compare_models


class State(BaseModel):
    """Poor man's state, don't even compare it to Composer's implementation.
    We typically do not include the dataloaders here, because they do not have
    `state` in a sense of how model or optimizers have. However, we can inherit
    `State` with `Serializable` and force the dataloaders to be included."""

    model: nn.Module = Field(default=None, description="Model.")

    criterion: nn.Module = Field(default=None, description="Loss function.")
    optimizer: torch.optim.Optimizer = Field(default=None, description="Optimizer.")
    scheduler: Union[torch.optim.lr_scheduler.LRScheduler, None] = Field(default=None, description="Scheduler.")

    epoch_index: int = Field(default=0, description="Current epoch index.")
    train_batch_index: int = Field(
        default=0, description="Current batch index and is only referring to the training batch index."
    )
    step_index: int = Field(
        default=0,
        description="We do not add prefix train because it is understood and implied that the step number is the train due to how many gradients been stepped. Current step index and is only referring to the training step index. What is the difference between step and batch? In general, they coincide for when the epoch number is 1, but after the first epoch, we usually reset the batch index to 0, while the step index keeps increasing to the next epoch.",
    )
    history: Dict[str, List[float]] = Field(default={}, description="History of metrics.")

    # FIXME: loosen `Vocabularies` and `Tokenizers` to `Any` for now as it is too strict.
    vocabulary: Any = Field(default=None, description="Vocabulary.")
    tokenizer: Any = Field(default=None, description="Tokenizer.")

    tokens_per_iter: int = Field(default=None, description="Tokens per iter/step.")

    def __eq__(self, other: object) -> bool:
        """Check if two State instances are equal."""
        assert isinstance(other, State), "Can only compare State instances."

        models_equal = compare_models(self.model, other.model)
        return models_equal

        # return (
        #     (self.model.state_dict() if self.model else None) == (other.model.state_dict() if other.model else None)
        #     and (self.criterion.state_dict() if self.criterion else None)
        #     == (other.criterion.state_dict() if other.criterion else None)
        #     and (self.optimizer.state_dict() if self.optimizer else None)
        #     == (other.optimizer.state_dict() if other.optimizer else None)
        #     and (self.scheduler.state_dict() if self.scheduler else None)
        #     == (other.scheduler.state_dict() if other.scheduler else None)
        #     and self.epoch_index == other.epoch_index
        #     and self.train_batch_index == other.train_batch_index
        # )

    class Config:
        """Pydantic config."""

        arbitrary_types_allowed = True

    @property
    def model_or_module(self) -> nn.Module:
        """The model not wrapped by :class:`DistributedDataParallel`."""
        if isinstance(self.model, DistributedDataParallel):
            return self.model.module  # type: ignore[no-any-return]
        return self.model

    def pretty_print(self) -> None:
        """Pretty print the config."""
        pprint(self)

    def save_snapshots(self, filepath: str) -> None:
        """Save the state dictionaries of the components to a file."""
        state = {
            "model": self.model_or_module.state_dict() if self.model_or_module else None,
            "criterion": self.criterion.state_dict() if self.criterion else None,
            "optimizer": self.optimizer.state_dict() if self.optimizer else None,
            "scheduler": self.scheduler.state_dict() if self.scheduler else None,
            "epoch_index": self.epoch_index,
            "train_batch_index": self.train_batch_index,
            "step_index": self.step_index,
            "history": self.history,
            "vocabulary": self.vocabulary,
            "tokenizer": self.tokenizer,
        }
        torch.save(state, filepath)

    @classmethod
    def load_snapshots(
        cls: Type[State],
        filepath: str,
        device: torch.device,
        *,
        model: nn.Module,
        criterion: nn.Module,
        optimizer: torch.optim.Optimizer,
        scheduler: torch.optim.lr_scheduler.LRScheduler,
    ) -> State:
        """Load state dictionaries from a file and return a new State instance."""
        state = torch.load(filepath, map_location=device)

        epoch_index = state["epoch_index"]
        train_batch_index = state["train_batch_index"]
        step_index = state["step_index"]
        history = state["history"]
        vocabulary = state["vocabulary"]
        tokenizer = state["tokenizer"]

        # Create a new instance of State with loaded state
        new_state = cls(
            model=model,
            criterion=criterion,
            optimizer=optimizer,
            scheduler=scheduler,
            epoch_index=epoch_index,
            train_batch_index=train_batch_index,
            step_index=step_index,
            history=history,
            vocabulary=vocabulary,
            tokenizer=tokenizer,
        )

        # Load state dicts into the model, criterion, etc., if they exist
        if new_state.model and "model" in state:
            new_state.model_or_module.load_state_dict(state["model"])
        if new_state.criterion and "criterion" in state:
            new_state.criterion.load_state_dict(state["criterion"])
        if new_state.optimizer and "optimizer" in state:
            new_state.optimizer.load_state_dict(state["optimizer"])
        if new_state.scheduler and "scheduler" in state:
            new_state.scheduler.load_state_dict(state["scheduler"])

        return new_state



"""LoRA: Low-Rank Adaptation of Large Language Models.

References
----------
[1] https://pytorch.org/torchtune/stable/tutorials/lora_finetune.html
"""


from __future__ import annotations

import math
from typing import List

import torch
from pydantic import BaseModel, Field
from torch import nn


class LoraConfig(BaseModel):
    r: int = Field(..., description="Lora attention dimension (the 'rank').")
    lora_alpha: int = Field(..., description="The alpha parameter for Lora scaling.")
    lora_dropout: float = Field(..., description="The dropout probability for Lora layers.")
    target_modules: List[str] = Field(
        default=None,
        description=(
            "The names of the modules to apply the adapter to. If specified, only the modules with the specified "
            "names will be replaced. When passing a string, a regex match will be performed. When passing a list of "
            "strings, either an exact match will be performed or it is checked if the name of the module ends with any "
            "of the passed strings. If specified as 'all-linear', all linear/Conv1D modules are chosen, excluding the "
            "output layer. If not specified, modules are chosen according to the model architecture. If the architecture "
            "is unknown, an error will be raised—manual specification of target modules is required in such cases."
        ),
    )
    linear_bias: bool = Field(default=True, description="To include linear bias or not.")
    modules_to_save: List[str] = Field(
        default=None,
        description=(
            """List of modules apart from adapter layers to be set as
               trainable and saved in the final checkpoint."""
        ),
    )


def _lora_a_init_params(x: nn.Linear) -> None:
    """
    Initialize LoRA A weight to Kaiming uniform.
    """
    nn.init.kaiming_uniform_(x.weight, a=math.sqrt(5))


def _lora_b_init_params(x: nn.Linear) -> None:
    """
    Initialize LoRA B weight to zeros.
    """
    nn.init.zeros_(x.weight)


class LoRALinear(nn.Module):
    def __init__(self, in_dim: int, out_dim: int, bias: bool, rank: int, alpha: float, dropout: float) -> None:
        super().__init__()

        # These are the weights from the original pretrained model
        self.linear = nn.Linear(in_dim, out_dim, bias=bias)  # weight shape=[out_dim, in_dim]

        # These are the new LoRA params. In general rank << in_dim, out_dim - do not put bias here
        self.lora_a = nn.Linear(in_features=in_dim, out_features=rank, bias=False)  # weight shape=[rank, in_dim]
        self.lora_b = nn.Linear(in_features=rank, out_features=out_dim, bias=False)  # weight shape=[out_dim, rank]

        self.rank = rank
        self.alpha = alpha
        self.dropout = nn.Dropout(p=dropout)

        self._init_weights()

    def _init_weights(self) -> None:
        """See https://github.com/microsoft/LoRA/blob/4c0333854cb905966f8cc4e9a74068c1e507c7b7/loralib/layers.py#L119."""

        _lora_a_init_params(self.lora_a)
        _lora_b_init_params(self.lora_b)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # This would be the output of the original model
        frozen_out = x @ self.linear.weight.T
        if self.linear.bias is not None:
            frozen_out += self.linear.bias

        # lora_a projects inputs down to the much smaller self.rank,
        # then lora_b projects back up to the output dimension
        x = self.dropout(x)
        lora_out = (x @ self.lora_a.weight.T) @ self.lora_b.weight.T  # x @ lora_a @ lora_b
        # Finally, scale by the alpha parameter (normalized by rank)
        # and add to the original model's outputs
        return frozen_out + (self.alpha / self.rank) * lora_out


def apply_lora_to_base_model(
    model: nn.Module, rank: int, alpha: float, dropout: float, target_modules: List[str] | None = None
) -> None:
    """Recursively apply LoRA to a model. Only supports applying on `nn.Linear` layers."""

    for module_name, module in model.named_children():
        if isinstance(module, nn.Linear):
            if target_modules is None or any(target in module_name for target in target_modules):
                setattr(
                    model,
                    module_name,
                    LoRALinear(
                        in_dim=module.in_features,
                        out_dim=module.out_features,
                        rank=rank,
                        alpha=alpha,
                        dropout=dropout,
                        bias=module.bias is not None,
                    ),
                )
        else:
            # Recursively apply LoRA to children modules
            apply_lora_to_base_model(
                model=module, rank=rank, alpha=alpha, dropout=dropout, target_modules=target_modules
            )



# mypy: disable-error-code="no-untyped-call, import-untyped"

from typing import List

import ipywidgets
from IPython.display import SVG, display


def create_tabbed_svg_viewer(image_paths: List[str], tab_titles: List[str]) -> ipywidgets.Tab:
    """
    Create a tabbed image viewer widget.

    Parameters
    ----------
    image_paths: List[str]
        A list containing the paths to the images to be displayed in the viewer.
    tab_titles : List[str]
        A list containing the titles for each tab in the viewer.

    Returns
    -------
    ipywidgets.Tab
        An IPython widget Tab instance with each tab containing an image from the
        provided paths and titled according to the `tab_titles` list.
    """
    if len(image_paths) != len(tab_titles):
        raise ValueError("image_paths and tab_titles must have the same length.")

    outputs = [ipywidgets.Output() for _ in image_paths]

    for output, image_path in zip(outputs, image_paths):
        with output:
            display(SVG(filename=image_path))

    tab = ipywidgets.Tab()
    tab.children = outputs

    for index, tab_title in enumerate(tab_titles):
        tab.set_title(index, str(tab_title))

    return tab



"""We use dataclass here for easy instantiating with hydra"""

from pydantic import BaseModel, Field
from torch import nn

from omnivault.transformer.modules.attention.base import Attention


class MultiHeadedAttentionConfig(BaseModel):
    attention: Attention
    d_model: int
    H: int
    dropout: float = 0.1

    class Config:
        """Pydantic config."""

        arbitrary_types_allowed = True


# TODO: add `field_validator` such that if `d_ff` is `None`, then `d_ff` is set to `4 * d_model`.
class PositionwiseFeedForwardConfig(BaseModel):
    d_model: int
    d_ff: int
    activation: nn.Module = Field(
        default=nn.GELU(approximate="tanh")
    )  # NOTE: https://github.com/facebookresearch/xformers/issues/759
    dropout: float = 0.1
    bias: bool = True

    class Config:
        """Pydantic config."""

        arbitrary_types_allowed = True


class AddNormConfig(BaseModel):
    feature_dim: int
    dropout: float


class DecoderBlockConfig(BaseModel):
    masked_self_attention_mha: MultiHeadedAttentionConfig
    feed_forward: PositionwiseFeedForwardConfig
    add_norm_1: AddNormConfig
    add_norm_2: AddNormConfig


class DecoderConfig(BaseModel):
    d_model: int
    vocab_size: int
    context_length: int  # NOTE: alias=max_seq_len,block_size
    num_decoder_blocks: int
    dropout: float
    decoder_block: DecoderBlockConfig



from typing import Callable

import torch
from torch import nn

from omnivault.transformer.modules.layers.normalization import LayerNorm


class AddNorm(nn.Module):
    """AddNorm is apt since the diagram uses Add + Norm.
    Some call it SubLayer connection (Harvard) some call it
    residual connection: x + dropout(sublayer(layernorm(x)))

    If stay true, then we apply residual then layer norm. So
    we adopt the d2l method."""

    def __init__(self, feature_dim: int, dropout: float) -> None:
        super().__init__()
        # fmt: off
        self.dropout    = nn.Dropout(p=dropout, inplace=False)
        self.layer_norm = LayerNorm(normalized_shape=feature_dim, eps=1e-5, elementwise_affine=True)
        # fmt: on

    def forward(self, x: torch.Tensor, sublayer: Callable[[torch.Tensor], torch.Tensor]) -> torch.Tensor:
        """The formulation is x + dropout(sublayer(layernorm(x)))."""
        # NOTE: GPT-2 should be from self.layer_norm(x + sublayer(self.dropout(x))) to x + self.dropout(sublayer(self.layer_norm(x)))
        output: torch.Tensor = x + self.dropout(sublayer(self.layer_norm(x)))
        return output



"""Base classes for decoders in transformer-like architectures. Template design pattern."""

from __future__ import annotations

from abc import ABC, abstractmethod

import torch
from torch import nn

from omnivault._types._alias import NotGiven
from omnivault._types._sentinel import NOT_GIVEN
from omnivault.transformer.config.decoder import DecoderConfig


class BaseDecoderBlock(ABC, nn.Module):
    """
    Abstract base class for a decoder block in a transformer-like architecture.
    """

    def __init__(self, config: DecoderConfig) -> None:
        super().__init__()
        self.config = config

    @abstractmethod
    def forward(
        self,
        z: torch.Tensor,  # that's tgt in torch code base
        *,  # force keyword only arguments to prevent errors
        encoder_hidden_states: torch.Tensor | NotGiven = NOT_GIVEN,  # that's memory in torch code base
        encoder_hidden_states_masks: torch.BoolTensor | NotGiven = NOT_GIVEN,  # that's memory_mask in torch code base
        target_masks: torch.BoolTensor | NotGiven = NOT_GIVEN,  # that's tgt_mask in torch code base
    ) -> torch.Tensor:
        """
        Performs one decoder *block* forward pass given final encoder hidden states, the previous block's output, and
        attention masks.

        N = batch size
        S = source sequence length
        T = target sequence length
        E = embedding dimensionality
        V = vocabulary size

        :param x: Previous decoder block's output. Shape: (N, T, E)
        :param encoder_hidden_states: The encoder's final (contextualized) token embeddings. Shape: (N, S, E)
        :param src_padding_mask: An attention mask to ignore pad-tokens in the source input. Shape (N, S)
        :param future_mask: An attention mask to ignore future-tokens in the target input. Shape (T, T)
        :return: Updated, contextualized token embeddings. Shape (N, T, E)
        """


class BaseDecoder(nn.Module, ABC):
    """
    Abstract base class for a decoder in a transformer-like architecture.
    """

    def __init__(
        self,
        config: DecoderConfig,
    ) -> None:
        super().__init__()
        self.config = config

    @abstractmethod
    def forward(
        self,
        input_tokens: torch.LongTensor,
        *,  # force keyword only arguments to prevent errors
        target_padding_masks: torch.BoolTensor | NotGiven = NOT_GIVEN,
        future_masks: torch.BoolTensor | NotGiven = NOT_GIVEN,
        encoder_hidden_states: torch.Tensor | NotGiven = NOT_GIVEN,  # that's memory in torch code base
        encoder_hidden_states_masks: torch.BoolTensor | NotGiven = NOT_GIVEN,  # that's memory_mask in torch code base
    ) -> torch.FloatTensor:
        ...

    def _init_weights(self, module: nn.Module) -> None:
        """Initializes weights of the given module using Xavier uniform initialization."""
        for p in module.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)



"""Module for creating PyTorch scheduler instances dynamically."""

from __future__ import annotations

from typing import Any, Callable, Dict, Literal, Type

import torch

from omnivault.utils.config_management.dynamic import DynamicClassFactory

RegisteredSchedulers = Literal[
    "torch.optim.lr_scheduler.StepLR",
    "torch.optim.lr_scheduler.CosineAnnealingLR",
    "torch.optim.lr_scheduler.LambdaLR",
    "torch.optim.lr_scheduler.CosineAnnealingWarmRestarts",
    "torch.optim.lr_scheduler.OneCycleLR",
]
SCHEDULER_REGISTRY: Dict[str, Type[SchedulerConfig]] = {}


def register_scheduler(name: str) -> Callable[[Type[SchedulerConfig]], Type[SchedulerConfig]]:
    def register_scheduler_cls(cls: Type[SchedulerConfig]) -> Type[SchedulerConfig]:
        if name in SCHEDULER_REGISTRY:
            raise ValueError(f"Cannot register duplicate scheduler {name}")
        if not issubclass(cls, SchedulerConfig):
            raise ValueError(f"Scheduler (name={name}, class={cls.__name__}) must extend SchedulerConfig")
        SCHEDULER_REGISTRY[name] = cls
        return cls

    return register_scheduler_cls


class SchedulerConfig(DynamicClassFactory[torch.optim.lr_scheduler.LRScheduler]):
    """
    Base class for creating PyTorch scheduler instances dynamically.

    This class extends `DynamicClassFactory` to specifically handle the
    instantiation of PyTorch scheduler classes based on provided configurations.

    Methods
    -------
    build(optimizer: torch.optim.Optimizer) -> torch.optim.lr_scheduler.LRScheduler
        Creates and returns a scheduler instance with the specified optimizer.
    """

    name: str

    def build(self, optimizer: torch.optim.Optimizer, **kwargs: Any) -> torch.optim.lr_scheduler.LRScheduler:
        """Builder method for creating a scheduler instance."""
        return self.create_instance(optimizer=optimizer, **kwargs)

    class Config:
        extra = "forbid"


@register_scheduler("torch.optim.lr_scheduler.StepLR")
class StepLRConfig(SchedulerConfig):
    step_size: int
    gamma: float = 0.1
    last_epoch: int = -1
    verbose: bool = False


@register_scheduler("torch.optim.lr_scheduler.CosineAnnealingLR")
class CosineAnnealingLRConfig(SchedulerConfig):
    T_max: int
    eta_min: float = 0
    last_epoch: int = -1
    verbose: bool = False


@register_scheduler("torch.optim.lr_scheduler.CosineAnnealingWarmRestarts")
class CosineAnnealingWarmRestartsConfig(SchedulerConfig):
    """See
    https://wandb.ai/wandb_fc/tips/reports/How-to-Properly-Use-PyTorch-s-CosineAnnealingWarmRestarts-Scheduler--VmlldzoyMTA3MjM2
    to see how to use this scheduler.
    """

    T_0: int
    T_mult: int = 1
    eta_min: float = 0
    last_epoch: int = -1
    verbose: bool = False


@register_scheduler("torch.optim.lr_scheduler.OneCycleLR")
class OneCycleLRConfig(SchedulerConfig):
    """You can use a utils function to retrieve the constructor args of a class
    dynamically.

    Example
    -------
    >>> from omnivault.utils.general.python_shenanigans import get_init_args
    >>> get_init_args(OneCycleLRConfig)
    """

    max_lr: float
    total_steps: int
    epochs: int
    steps_per_epoch: int
    pct_start: float = 0.3
    anneal_strategy: Literal["cos", "linear"] = "cos"
    cycle_momentum: bool = True
    base_momentum: float = 0.85
    max_momentum: float = 0.95
    div_factor: float = 25.0
    final_div_factor: float = 10000.0
    three_phase: bool = False
    last_epoch: int = -1
    verbose: bool = False


@register_scheduler("torch.optim.lr_scheduler.LambdaLR")
class LambdaLRConfig(SchedulerConfig):
    # The user must provide a lambda function for the scheduler
    lr_lambda: Callable[[int], float]  # we know this lr_lambda maps int (epoch) to float (some multiplier)



from typing import Callable

import torch
from torch import nn


class ResidualBlock(nn.Module):
    def forward(
        self,
        x: torch.Tensor,
        sublayer: Callable[[torch.Tensor], torch.Tensor],
    ) -> torch.Tensor:
        return x + sublayer(x)



from __future__ import annotations

from enum import IntEnum


class MemoryUnit(IntEnum):
    BIT = 2**0
    BYTE = 2**3

    # Binary Units In Bytes
    KiB = 2**10  # 1024 BYTES -> 1024 * 8 = 8192 BITs
    MiB = 2**20  # 1024 * KiB
    GiB = 2**30  # 1024 * MiB
    TiB = 2**40  # 1024 * GiB
    PiB = 2**50  # 1024 * TiB

    # Decimal Units
    KB = 10**3  # 1000 * BYTE
    MB = 10**6  # 1000 * KB
    GB = 10**9  # 1000 * MB
    TB = 10**12  # 1000 * GB
    PB = 10**15  # 1000 * TB

    @staticmethod
    def convert(value_in_bytes: int, from_unit: MemoryUnit, to_unit: MemoryUnit) -> float:
        """Converts memory unit from one to another.

        Example
        -------
        >>> value = 20480000
        >>> from_unit = MemoryUnit.BYTE
        >>> to_unit = MemoryUnit.GiB
        >>> result = MemoryUnit.convert(value, from_unit, to_unit)
        0.152587890625 # 20480000 bytes is 0.152587890625 GiB
        """
        total_bytes = value_in_bytes * from_unit
        return total_bytes / to_unit



"""
This file contains aliases for common types.
"""
from __future__ import annotations

from typing_extensions import TypeAlias

from omnivault._types._sentinel import _Missing, _NotGiven, _Omit

NonNegativeInt: TypeAlias = int
PositiveInt: TypeAlias = int
Loss: TypeAlias = float
Accuracy: TypeAlias = float
Token: TypeAlias = str
NotGiven: TypeAlias = _NotGiven
Missing: TypeAlias = _Missing
Omit: TypeAlias = _Omit



# NOTE: unit-tested
"""Even though the module is called pairwise, the functions are not pairwise.
This will be implemented in due time, so for now it just expects two 1D arrays."""
from typing import Any

import numpy as np
from numpy.typing import NDArray


def manhattan_distance(x_1: NDArray[np.floating[Any]], x_2: NDArray[np.floating[Any]]) -> float:
    r"""
    Calculate the Manhattan Distance between two data points.

    .. note::
        This is only 1D implementation of the Manhattan Distance and therefore
        is not pairwise. For example, if you have two data points X and Y
        with X = [[1, 2], [3, 4]] and Y = [[5, 6], [7, 8]], a pairwise Manhattan
        Distance would calculate the distance between the first data point in X
        and the first data point in Y, the distance between the first data point
        in X and the second data point in Y, the distance between the second data
        point in X and the first data point in Y, and the distance between the
        second data point in X and the second data point in Y.

    The Manhattan Distance is defined as the sum of the absolute differences
    between the coordinates of two vectors and it extends to N-dimensional
    Euclidean Space. It is formally defined as:

    .. math::

        L1(a, b) = \sum_{i=1}^{n} |a_i - b_i|

    Parameters
    ----------
    x_1 : NDArray[np.floating[Any]]
        A numpy array representing the first data point.
    x_2 : NDArray[np.floating[Any]]
        A numpy array representing the second data point.

    Returns
    -------
    float
        The Manhattan distance between the two data points.

    Examples
    --------
    >>> x_1 = np.array([1, 2, 3])
    >>> x_2 = np.array([2, 3, 5])
    >>> manhattan_distance(x_1, x_2)
    4
    """
    _manhattan_distance = np.sum(np.abs(x_1 - x_2))
    return float(_manhattan_distance)


def euclidean_distance(x_1: NDArray[np.floating[Any]], x_2: NDArray[np.floating[Any]], squared: bool = False) -> float:
    r"""
    Euclidean Distance measures the length of the line segment bewteen two points
    in the Euclidean space. This is also referred to as L2 distance or L2 vector norm.

    .. math::

        L2(a, b) = \sqrt{\sum_{i=1}^{n} (a_i - b_i)^2}


    This function generalizes to any number of dimensions:

    For example, for two points in 3D space (x1, y1, z1) and (x2, y2, z2),
    the Euclidean distance is:

    .. math::

        \sqrt{(x1-x2)^2 + (y1-y2)^2 + (z1-z2)^2}

    Parameters
    ----------
    x_1 : NDArray[np.floating[Any]]
        An N-dimensional numpy array representing the first data point.
    x_2 : NDArray[np.floating[Any]]
        An N-dimensional numpy array representing the second data point.
    squared : bool, optional
        If True, return the squared Euclidean distance.

    Returns
    -------
    float
        The Euclidean distance (or squared Euclidean distance, if `squared` is True)
        between the two data points.

    Examples
    --------
    >>> x_1 = np.array([1, 2, 3])
    >>> x_2 = np.array([2, 3, 5])
    >>> euclidean_distance(x_1, x_2)
    2.23606797749979
    >>> euclidean_distance(x_1, x_2, squared=True)
    5.0
    """
    _euclidean_distance = np.sum(np.square(x_1 - x_2)) if not squared else np.sqrt(np.sum(np.square(x_1 - x_2)))
    return float(_euclidean_distance)


def cosine_similarity(x_1: NDArray[np.floating[Any]], x_2: NDArray[np.floating[Any]]) -> float:
    r"""
    Calculate the cosine similarity between two vectors.

    The cosine similarity measures the cosine of the angle between two vectors
    projected in a multi-dimensional space. The cosine similarity is particularly
    used in high-dimensional positive spaces. It is defined as:

    .. math::

        \text{cosine similarity} = \frac{\mathbf{x}_1 \cdot \mathbf{x}_2}
                                        {||\mathbf{x}_1|| \cdot ||\mathbf{x}_2||}

    where \(\mathbf{x}_1\) and \(\mathbf{x}_2\) are vector representations.

    Parameters
    ----------
    x_1 : NDArray[np.floating[Any]]
        An N-dimensional numpy array representing the first vector.
    x_2 : NDArray[np.floating[Any]]
        An N-dimensional numpy array representing the second vector.

    Returns
    -------
    float
        The cosine similarity between the two vectors.

    Raises
    ------
    ValueError
        If either of the vectors is zero-length.
    AssertionError
        If the norm of x_1 as calculated does not match the Euclidean distance
        from x_1 to the origin.

    Examples
    --------
    >>> x_1 = np.array([1, 2, 3])
    >>> x_2 = np.array([2, 3, 4])
    >>> cosine_similarity(x_1, x_2)
    0.9925833339709303
    """
    if np.linalg.norm(x_1) == 0 or np.linalg.norm(x_2) == 0:
        raise ValueError("Cosine similarity is undefined for zero-length vectors.")

    numerator = np.dot(x_1, x_2)
    origin = np.zeros(shape=(x_1.shape))  # origin is a vector of zeros
    norm_x1 = np.linalg.norm(x_1)
    norm_x2 = np.linalg.norm(x_2)

    np.testing.assert_allclose(norm_x1, euclidean_distance(x_1, origin, squared=True))

    denominator = norm_x1 * norm_x2
    _cosine_similarity = numerator / denominator
    return float(_cosine_similarity)


def cosine_distance(x_1: NDArray[np.floating[Any]], x_2: NDArray[np.floating[Any]]) -> float:
    r"""
    Calculate the cosine distance between two vectors.

    Cosine distance is defined as 1 minus the cosine similarity:

    .. math::

        \text{cosine distance} = 1 - \text{cosine similarity}(\mathbf{x}_1, \mathbf{x}_2)

    Parameters
    ----------
    x_1 : NDArray[np.floating[Any]]
        An N-dimensional numpy array representing the first vector.
    x_2 : NDArray[np.floating[Any]]
        An N-dimensional numpy array representing the second vector.

    Returns
    -------
    float
        The cosine distance between the two vectors.

    Examples
    --------
    >>> x_1 = np.array([1, 2, 3])
    >>> x_2 = np.array([2, 3, 4])
    >>> cosine_distance(x_1, x_2)
    0.007416666029069703
    """
    return 1 - cosine_similarity(x_1, x_2)



def noam_lr_decay(step: int, *, d_model: int, warmup_steps: int) -> float:
    """
    Calculate the learning rate based on the current step, model dimensionality,
    and the number of warmup steps.

    Parameters
    ----------
    step : int
        The current step in the training process.
    d_model : int
        The dimensionality of the model.
    warmup_steps : int
        The number of steps during which the learning rate is increased linearly.
        After reaching `warmup_steps`, the learning rate decays proportionally to the
        inverse square root of the step number.

    Returns:
        float: The calculated learning rate.
    """

    return float(d_model ** (-0.5) * min((step + 1) ** (-0.5), (step + 1) * warmup_steps ** (-1.5)))



"""
Reference:
- https://peps.python.org/pep-0661/
- https://github.com/taleinat/python-stdlib-sentinels/blob/main/sentinels/sentinels.py
"""

from __future__ import annotations

import sys as _sys
from threading import Lock as _Lock
from types import FrameType
from typing import Tuple, Type

__all__ = ["Sentinel"]

# Design and implementation decisions:
#
# The first implementations created a dedicated class for each instance.
# However, once it was decided to use Sentinel for type signatures, there
# was no longer a need for a dedicated class for each sentinel value on order
# to enable strict type signatures.  Since class objects consume a relatively
# large amount of memory, the implementation was changed to avoid this.
#
# With this change, the mechanism used for unpickling/copying objects needed
# to be changed too, since we could no longer count on each dedicated class
# simply returning its singleton instance as before.  __reduce__ can return
# a string, upon which an attribute with that name is looked up in the module
# and returned.  However, that would have meant that pickling/copying support
# would depend on the "name" argument being exactly the name of the variable
# used in the module, and simply wouldn't work for sentinels created in
# functions/methods.  Instead, a registry for sentinels was added, where all
# sentinel objects are stored keyed by their name + module name.  This is used
# to look up existing sentinels both during normal object creation and during
# copying/unpickling.


class Sentinel:
    """Create a unique sentinel object.

    *name* should be the fully-qualified name of the variable to which the
    return value shall be assigned.

    *repr*, if supplied, will be used for the repr of the sentinel object.
    If not provided, "<name>" will be used (with any leading class names
    removed).

    *module_name*, if supplied, will be used instead of inspecting the call
    stack to find the name of the module from which
    """

    _name: str
    _repr: str
    _module_name: str

    def __new__(  # noqa: PYI034
        cls: Type[Sentinel],
        name: str,
        repr: str | None = None,
        module_name: str | None = None,
    ) -> Sentinel:
        name = str(name)
        repr = str(repr) if repr else f'<{name.split(".")[-1]}>'
        if not module_name:
            parent_frame = _get_parent_frame()
            module_name = parent_frame.f_globals.get("__name__", "__main__") if parent_frame is not None else __name__

        # Include the class's module and fully qualified name in the
        # registry key to support sub-classing.
        registry_key = _sys.intern(f"{cls.__module__}-{cls.__qualname__}-{module_name}-{name}")
        sentinel = _registry.get(registry_key, None)
        if sentinel is not None:
            return sentinel
        sentinel = super().__new__(cls)
        sentinel._name = name
        sentinel._repr = repr
        sentinel._module_name = module_name
        with _lock:
            return _registry.setdefault(registry_key, sentinel)

    def __repr__(self) -> str:
        return self._repr

    def __reduce__(self) -> Tuple[Type[Sentinel], Tuple[str, str, str]]:
        return (
            self.__class__,
            (
                self._name,
                self._repr,
                self._module_name,
            ),
        )


_lock = _Lock()
_registry: dict[str, Sentinel] = {}


# The following implementation attempts to support Python
# implementations which don't support sys._getframe(2), such as
# Jython and IronPython.
#
# The version added to the stdlib may simply return sys._getframe(2),
# without the fallbacks.
#
# For reference, see the implementation of namedtuple:
# https://github.com/python/cpython/blob/67444902a0f10419a557d0a2d3b8675c31b075a9/Lib/collections/__init__.py#L503
def _get_parent_frame() -> FrameType | None:
    """Return the frame object for the caller's parent stack frame."""
    try:
        # Two frames up = the parent of the function which called this.
        return _sys._getframe(2)
    except (AttributeError, ValueError):
        global _get_parent_frame

        def _get_parent_frame() -> FrameType | None:
            """Return the frame object for the caller's parent stack frame."""
            try:
                raise Exception
            except Exception:
                try:
                    return _sys.exc_info()[2].tb_frame.f_back.f_back  # type: ignore[union-attr]
                except Exception:
                    global _get_parent_frame

                    def _get_parent_frame() -> None:
                        """Return the frame object for the caller's parent stack frame."""
                        return None

                    return _get_parent_frame()

        return _get_parent_frame()



import multiprocessing
import os
import socket
import warnings

import torch
import torch.distributed


def distributed_available() -> None:
    """
    Check if distributed training is available and raises error
    otherwise.
    """
    if not torch.distributed.is_available():
        raise RuntimeError(
            "Distributed training is not available. "
            "Please check that PyTorch is built with distributed support and "
            "the necessary distributed packages are installed."
        )


def distributed_initialized() -> None:
    """
    Check if distributed training is initialized and raises error otherwise.
    """
    if not torch.distributed.is_initialized():
        raise RuntimeError(
            "Distributed training is not initialized. "
            "Please ensure that `torch.distributed.init_process_group()` is called "
            "before using distributed training functions."
        )


def get_world_size() -> int:
    """
    Return the total number of processes participating in the distributed training.

    This function determines the global size of the process group, which represents
    the total number of processes across all nodes involved in the distributed
    training.

    Example
    -------
    -   If the training is running on a single node with 4 processes, the function
        will return 4.
    -   If the training is running on 2 nodes with 4 processes per node, the
        function will return 8 (2 nodes * 4 processes per node).

    Note
    ----
    If we are using `torchrun`, then we can also obtain the world size using
    `os.environ["WORLD_SIZE"]`.
    """
    distributed_available()
    distributed_initialized()
    return torch.distributed.get_world_size()


def get_local_world_size() -> int:
    """
    Return the number of processes per node participating in the distributed training.

    This function determines the local size of the process group, which represents
    the number of processes running on the same node.

    Example
    -------
    -   If the training is running on a single node with 4 processes, the function
        will return 4.
    -   If the training is running on 2 nodes with 4 processes per node, the
        function will return 4 (4 processes per node).

    Note
    ----
    If we are using `torchrun`, then we can also obtain the local world size using
    `os.environ["LOCAL_WORLD_SIZE"]`.
    """
    if torch.cuda.is_available():
        return torch.cuda.device_count()
    else:
        warnings.warn("CUDA is not available. Returning total number of CPU cores.", stacklevel=2)
        return multiprocessing.cpu_count()


def get_global_rank() -> int:
    """
    Return the rank of the current process across all nodes.

    This function determines the global rank of the current process, which
    represents the index of the current process across all nodes involved in
    the distributed training.

    Example
    -------
    -   If the training is running on a single node with 4 processes, the function
        will return the rank of the process on that node.
    -   If the training is running on 2 nodes with 4 processes per node, the
        function will return the rank of the process across all nodes.

    Note
    ----
    If we are using `torchrun`, then we can also obtain the global rank using
    `os.environ["RANK"]`.
    """
    distributed_available()
    distributed_initialized()
    return torch.distributed.get_rank()


def get_local_rank() -> int:
    """
    Return the rank of the current process on the current node.

    This function determines the local rank of the current process, which
    represents the index of the current process on the current node.

    Example
    -------
    -   If the training is running on a single node with 4 processes, the function
        will return the rank of the process on that node.
    -   If the training is running on 2 nodes with 4 processes per node, the
        function will return the rank of the process on the current node.

    Note
    ----
    If we are using `torchrun`, then we can also obtain the local rank using
    `os.environ["LOCAL_RANK"]`.
    """
    distributed_available()
    distributed_initialized()
    return get_global_rank() % get_local_world_size()


def is_master_rank() -> bool:
    """
    Check if the current process is the master process.

    The master process is the process with rank 0 across all nodes involved
    in the distributed training.
    """
    return get_global_rank() == 0


def synchronize_and_barrier() -> None:
    if torch.cuda.is_available():
        torch.cuda.synchronize()
    torch.distributed.barrier()


def teardown() -> None:
    torch.distributed.destroy_process_group()


def is_free_port(port: int) -> bool:
    """Checks if a given port is free on all relevant interfaces.

    References
    ----------
    -   https://github.com/serend1p1ty/core-pytorch-utils
    """
    ips = ["localhost", "127.0.0.1"]
    try:
        ips.extend(socket.gethostbyname_ex(socket.gethostname())[-1])
    except socket.gaierror:
        warnings.warn("Failed to get all IPs for the current host. Only using localhost", stacklevel=2)

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return all(sock.connect_ex((ip, port)) != 0 for ip in ips)


def find_free_port() -> int:
    """Finds a free port by binding to port 0 and then checking if it's truly free."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.bind(("", 0))
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        port: int = sock.getsockname()[1]
        if is_free_port(port):
            return port
        else:
            return find_free_port()  # Recursively search if the first found port is not free


def get_hostname() -> str:
    """Get the hostname of the machine."""
    return socket.gethostname()


def get_process_id() -> int:
    """Get the process id of the current process."""
    return os.getpid()



import torch

from omnivault.constants.memory import MemoryUnit


def create_tensor_of_vram_size(dtype: torch.dtype, vram_size_in_bytes: int) -> torch.Tensor:
    """Create a tensor of the specified dtype that fits in the VRAM size."""

    # 1. For example, if `dtype` is `torch.float32` then `bytes_per_element` is 4.
    #    It returns the bytes needed to store a single element of the tensor.
    if dtype.is_floating_point:
        bytes_per_element = torch.finfo(dtype).bits // MemoryUnit.BYTE
        assert (
            bytes_per_element == torch.tensor([], dtype=dtype).element_size()
        )  # TODO: may be inefficient adding this assertion
    else:
        bytes_per_element = torch.iinfo(dtype).bits // MemoryUnit.BYTE

    # 2. Simple math, we need to find the number of elements required to
    #    "consume" the target vram. For example, if we want to consume 10MB of
    #    vram and each element is 4 bytes (float32), then we need ~2.5 million
    #    elements derived from 10MB / 4 bytes per element.
    total_elements_needed = int(vram_size_in_bytes / bytes_per_element)

    # 3. Create 1D tensor with the required number of elements.
    tensor = torch.empty(total_elements_needed, dtype=dtype)
    assert tensor.size() == (
        total_elements_needed,
    ), f"Expected tensor size {total_elements_needed} but got {tensor.size()}."

    return tensor



from __future__ import annotations

from typing import Any, Callable, Dict, Iterator, List, Literal, Tuple, Type

import torch
from torch import nn

from omnivault.utils.config_management.dynamic import DynamicClassFactory

RegisteredOptimizers = Literal["torch.optim.Adam", "torch.optim.AdamW", "torch.optim.SGD"]
OPTIMIZER_REGISTRY: Dict[RegisteredOptimizers, Type[OptimizerConfig]] = {}


def register_optimizer(name: RegisteredOptimizers) -> Callable[[Type[OptimizerConfig]], Type[OptimizerConfig]]:
    """
    Decorator factory for registering optimizer configurations.

    Parameters
    ----------
    name : RegisteredOptimizers
        The name to register the class under; if not provided, the class's 'name' attribute is used.

    Returns
    -------
    Callable[[Type[OptimizerConfig]], Type[OptimizerConfig]]
        A decorator that registers the optimizer configuration class.
    """

    def register_optimizer_cls(cls: Type[OptimizerConfig]) -> Type[OptimizerConfig]:
        if name in OPTIMIZER_REGISTRY:
            raise ValueError(f"Cannot register duplicate optimizer {name}")

        if not issubclass(cls, OptimizerConfig):
            raise ValueError(f"Optimizer (name={name}, class={cls.__name__}) must extend OptimizerConfig")

        OPTIMIZER_REGISTRY[name] = cls
        return cls

    return register_optimizer_cls


# NOTE: remember to give generic class a type here, which is
# aptly `torch.optim.Optimizer` in this case.
class OptimizerConfig(DynamicClassFactory[torch.optim.Optimizer]):
    """
    Configuration for creating PyTorch optimizer instances dynamically.

    This class extends `DynamicClassFactory` to specifically handle the
    instantiation of PyTorch optimizer classes based on provided configurations.
    The primary use case is to create optimizers with different settings
    for experimenting with model training.

    Parameters
    ----------
    name : str
        The fully qualified class name of the optimizer to instantiate.
        Defaults to 'torch.optim.Adam'.
    lr : float
        Learning rate for the optimizer. Defaults to 0.2.
    betas : Tuple[float, float]
        Coefficients used for computing running averages of gradient and its square.
        Defaults to (0.9, 0.98).
    eps : float
        Term added to the denominator to improve numerical stability.
        Defaults to 1e-9.

    Methods
    -------
    build(params: nn.ParameterList) -> torch.optim.Optimizer
        Creates and returns an optimizer instance with the specified parameters.

    Examples
    --------
    >>> import torch
    >>> from torch import nn
    >>> from omnivault.utils.config_management.dynamic import DynamicClassFactory
    >>> parameters = nn.ParameterList([nn.Parameter(torch.randn(2, 2, requires_grad=True))])
    >>> optimizer_config = OptimizerConfig(name="torch.optim.Adam", lr=0.1, betas=(0.9, 0.999), eps=1e-8)
    >>> optimizer = optimizer_config.build(params=parameters)

    Below is an example that works but not recommended:
    >>> optimizer_config = OptimizerConfig(name="torch.optim.Adam", lr=0.1)
    >>> optimizer = optimizer_config.build(params=parameters, betas=(0.9, 0.999), eps=1e-8)

    This works but it is harder to maintain as we are passing in the "config" in
    `build` method instead of the constructor. The recommended way is to pass in
    the config in the constructor and only pass in the model params in the `build`
    method.
    """

    name: str
    lr: float  # assume all optimizers have this parameter

    # FIXME: how do we loosen this params type?
    def build(
        self,
        *,
        params: List[Dict[Literal["params", "weight_decay"], List[torch.nn.Parameter] | float]]
        | nn.ParameterList
        | Iterator[nn.Parameter],
        **kwargs: Any,
    ) -> torch.optim.Optimizer:
        """Builder method for creating an optimizer instance."""
        return self.create_instance(params=params, **kwargs)

    class Config:
        """Pydantic configuration for `OptimizerConfig`."""

        extra = "forbid"


@register_optimizer(name="torch.optim.Adam")
class AdamConfig(OptimizerConfig):
    betas: Tuple[float, float] = (0.9, 0.98)
    eps: float = 1e-9
    weight_decay: float = 0.0


@register_optimizer(name="torch.optim.AdamW")
class AdamWConfig(OptimizerConfig):
    betas: Tuple[float, float] = (0.9, 0.999)
    eps: float = 1e-8
    weight_decay: float = 1e-2
    amsgrad: bool = False


@register_optimizer(name="torch.optim.SGD")
class SGDConfig(OptimizerConfig):
    momentum: float = 0.0
    dampening: float = 0.0
    weight_decay: float = 0.0


# At the end of this script, `OPTIMIZER_REGISTRY` will be populated with
# the following:
# OPTIMIZER_REGISTRY = {
#     "torch.optim.Adam": AdamConfig,
#     "torch.optim.SGD": SGDConfig,
# }
# This is because the `register_optimizer` decorator is called during
# runtime.



from __future__ import annotations

from typing import Union

from pydantic import BaseModel, Field
from rich.pretty import pprint

from omnivault._types._alias import Missing
from omnivault._types._sentinel import MISSING
from omnivault.transformer.config.constants import MaybeConstant
from omnivault.transformer.config.criterion import CriterionConfig
from omnivault.transformer.config.data import DataConfig
from omnivault.transformer.config.decoder import DecoderConfig
from omnivault.transformer.config.distributed import DistributedConfig
from omnivault.transformer.config.generator import GeneratorConfig
from omnivault.transformer.config.global_ import MaybeGlobal
from omnivault.transformer.config.logger import LoggerConfig
from omnivault.transformer.config.optim import OptimizerConfig
from omnivault.transformer.config.scheduler import SchedulerConfig
from omnivault.transformer.config.trainer import TrainerConfig


class Composer(BaseModel):  # TODO: add generic subclassing - see if got time lols
    constants: MaybeConstant = Field(default_factory=MaybeConstant)
    logger: LoggerConfig = Field(default_factory=LoggerConfig)
    global_: MaybeGlobal = Field(default_factory=MaybeGlobal)
    data: DataConfig = Field(default_factory=DataConfig)
    model: Union[DecoderConfig, Missing] = Field(default=MISSING, description="The model config.")
    optimizer: Union[OptimizerConfig, Missing] = Field(default=MISSING, description="The optimizer config.")
    criterion: Union[CriterionConfig, Missing] = Field(default=MISSING, description="The criterion config.")
    scheduler: Union[SchedulerConfig, Missing] = Field(default=MISSING, description="The scheduler config.")
    trainer: TrainerConfig = Field(default_factory=TrainerConfig)
    generator: GeneratorConfig = Field(default_factory=GeneratorConfig)
    distributed: DistributedConfig = Field(default_factory=DistributedConfig)

    class Config:
        """Pydantic config."""

        arbitrary_types_allowed = True

    def pretty_print(self) -> None:
        """Pretty print the config."""
        pprint(self)
